from eth_keys.backends.base import (
    BaseECCBackend,
)
from eth_keys.datatypes import (
    BaseSignature,
    NonRecoverableSignature,
    PrivateKey,
    PublicKey,
    Signature,
)

from .ecdsa import (
    compress_public_key,
    decompress_public_key,
    ecdsa_raw_recover,
    ecdsa_raw_sign,
    ecdsa_raw_verify,
    private_key_to_public_key,
)


class NativeECCBackend(BaseECCBackend):
    def ecdsa_sign(self, msg_hash: bytes, private_key: PrivateKey) -> Signature:
        signature_vrs = ecdsa_raw_sign(msg_hash, private_key.to_bytes())
        signature = Signature(vrs=signature_vrs, backend=self)
        return signature

    def ecdsa_sign_non_recoverable(
        self, msg_hash: bytes, private_key: PrivateKey
    ) -> NonRecoverableSignature:
        _, signature_r, signature_s = ecdsa_raw_sign(msg_hash, private_key.to_bytes())
        signature = NonRecoverableSignature(rs=(signature_r, signature_s), backend=self)
        return signature

    def ecdsa_verify(
        self, msg_hash: bytes, signature: BaseSignature, public_key: PublicKey
    ) -> bool:
        return ecdsa_raw_verify(msg_hash, signature.rs, public_key.to_bytes())

    def ecdsa_recover(self, msg_hash: bytes, signature: Signature) -> PublicKey:
        public_key_bytes = ecdsa_raw_recover(msg_hash, signature.vrs)
        public_key = PublicKey(public_key_bytes, backend=self)
        return public_key

    def private_key_to_public_key(self, private_key: PrivateKey) -> PublicKey:
        public_key_bytes = private_key_to_public_key(private_key.to_bytes())
        public_key = PublicKey(public_key_bytes, backend=self)
        return public_key

    def decompress_public_key_bytes(self, compressed_public_key_bytes: bytes) -> bytes:
        return decompress_public_key(compressed_public_key_bytes)

    def compress_public_key_bytes(self, uncompressed_public_key_bytes: bytes) -> bytes:
        return compress_public_key(uncompressed_public_key_bytes)



import pytest

from eth_keys import (
    KeyAPI,
)
from eth_keys.backends import (
    NativeECCBackend,
)


@pytest.fixture(autouse=True)
def native_backend_env_var(monkeypatch):
    monkeypatch.setenv("ECC_BACKEND_CLASS", "eth_keys.backends.native.NativeECCBackend")


@pytest.mark.parametrize(
    "backend",
    (
        None,
        NativeECCBackend(),
        NativeECCBackend,
        "eth_keys.backends.NativeECCBackend",
        "eth_keys.backends.native.NativeECCBackend",
    ),
)
def test_supported_backend_formats(backend):
    keys = KeyAPI(backend=backend)
    assert isinstance(keys.backend, NativeECCBackend)



from eth_utils import (
    big_endian_to_int,
)

from eth_keys.datatypes import (
    BaseSignature,
    NonRecoverableSignature,
    PrivateKey,
    PublicKey,
    Signature,
)
from eth_keys.exceptions import (
    BadSignature,
)
from eth_keys.utils import (
    der,
)
from eth_keys.utils.numeric import (
    coerce_low_s,
)
from eth_keys.validation import (
    validate_uncompressed_public_key_bytes,
)

from .base import (
    BaseECCBackend,
)


def is_coincurve_available() -> bool:
    try:
        import coincurve  # noqa: F401
    except ImportError:
        return False
    else:
        return True


class CoinCurveECCBackend(BaseECCBackend):
    def __init__(self) -> None:
        try:
            import coincurve
        except ImportError:
            raise ImportError(
                "The CoinCurveECCBackend requires the coincurve "
                "library which is not available for import."
            )
        self.keys = coincurve.keys
        self.ecdsa = coincurve.ecdsa
        super().__init__()

    def ecdsa_sign(self, msg_hash: bytes, private_key: PrivateKey) -> Signature:
        private_key_bytes = private_key.to_bytes()
        signature_bytes = self.keys.PrivateKey(private_key_bytes).sign_recoverable(
            msg_hash,
            hasher=None,
        )
        signature = Signature(signature_bytes, backend=self)
        return signature

    def ecdsa_sign_non_recoverable(
        self, msg_hash: bytes, private_key: PrivateKey
    ) -> NonRecoverableSignature:
        private_key_bytes = private_key.to_bytes()

        der_encoded_signature = self.keys.PrivateKey(private_key_bytes).sign(
            msg_hash,
            hasher=None,
        )
        rs = der.two_int_sequence_decoder(der_encoded_signature)

        signature = NonRecoverableSignature(rs=rs, backend=self)
        return signature

    def ecdsa_verify(
        self, msg_hash: bytes, signature: BaseSignature, public_key: PublicKey
    ) -> bool:
        # coincurve rejects signatures with a high s,
        # so convert to the equivalent low s form
        low_s = coerce_low_s(signature.s)
        der_encoded_signature = der.two_int_sequence_encoder(signature.r, low_s)
        coincurve_public_key = self.keys.PublicKey(b"\x04" + public_key.to_bytes())
        return coincurve_public_key.verify(
            der_encoded_signature,
            msg_hash,
            hasher=None,
        )

    def ecdsa_recover(self, msg_hash: bytes, signature: Signature) -> PublicKey:
        signature_bytes = signature.to_bytes()
        try:
            public_key_bytes = self.keys.PublicKey.from_signature_and_message(
                signature_bytes,
                msg_hash,
                hasher=None,
            ).format(compressed=False)[1:]
        except Exception as err:
            raise BadSignature(str(err))
        public_key = PublicKey(public_key_bytes, backend=self)
        return public_key

    def private_key_to_public_key(self, private_key: PrivateKey) -> PublicKey:
        public_key_bytes = self.keys.PrivateKey(
            private_key.to_bytes()
        ).public_key.format(compressed=False,)[1:]
        return PublicKey(public_key_bytes, backend=self)

    def decompress_public_key_bytes(self, compressed_public_key_bytes: bytes) -> bytes:
        public_key = self.keys.PublicKey(compressed_public_key_bytes)
        return public_key.format(compressed=False)[1:]

    def compress_public_key_bytes(self, uncompressed_public_key_bytes: bytes) -> bytes:
        validate_uncompressed_public_key_bytes(uncompressed_public_key_bytes)
        point = (
            big_endian_to_int(uncompressed_public_key_bytes[:32]),
            big_endian_to_int(uncompressed_public_key_bytes[32:]),
        )
        public_key = self.keys.PublicKey.from_point(*point)
        return public_key.format(compressed=True)



import pytest

# Change COLLECT_TYPE_INFO to True here, and run the tests with python2.7 to get a type
# info dump that can later be fed to pyannotate to generate type annotation comments.
COLLECT_TYPE_INFO = False

if COLLECT_TYPE_INFO:
    from pyannotate_runtime import (
        collect_types,
    )

    @pytest.fixture(autouse=True)
    def collect_types_fixture():
        collect_types.resume()
        yield
        collect_types.pause()

    def pytest_sessionstart(session):
        collect_types.init_types_collection()

    def pytest_sessionfinish(session, exitstatus):
        collect_types.dump_stats("type_info.json")



from eth_utils import (
    int_to_big_endian,
)
from hypothesis import (
    strategies as st,
)

from eth_keys.constants import (
    SECPK1_N,
)
from eth_keys.utils.padding import (
    pad32,
)

private_key_st = (
    st.integers(min_value=1, max_value=SECPK1_N)
    .map(
        int_to_big_endian,
    )
    .map(pad32)
)


message_hash_st = st.binary(min_size=32, max_size=32)
signature_st = st.binary(min_size=65, max_size=65)



# This exposes the eth-utils exception for backwards compatibility,
# for any library that catches eth_keys.exceptions.ValidationError
from eth_utils import ValidationError  # noqa: F401


class BadSignature(Exception):
    pass



from eth_utils import (
    ValidationError,
    decode_hex,
    encode_hex,
    is_canonical_address,
    is_checksum_address,
    is_normalized_address,
    is_same_address,
    keccak,
)
import pytest

from eth_keys import (
    KeyAPI,
)
from eth_keys.backends import (
    NativeECCBackend,
)
from eth_keys.exceptions import (
    ValidationError as EthKeysValidationErrorCopy,
)

MSG = b"message"
MSGHASH = keccak(MSG)


@pytest.fixture
def key_api():
    return KeyAPI(backend=NativeECCBackend())


PK_BYTES = decode_hex(
    "0x58d23b55bc9cdce1f18c2500f40ff4ab7245df9a89505e9b1fa4851f623d241d"
)
ADDRESS = "0xdc544d1aa88ff8bbd2f2aec754b1f1e99e1812fd"


@pytest.fixture
def private_key(key_api):
    return key_api.PrivateKey(PK_BYTES)


def test_signing_from_private_key_obj(key_api, private_key):
    signature = private_key.sign_msg(MSG)

    assert key_api.ecdsa_verify(MSGHASH, signature, private_key.public_key)


def test_signing_non_recoverable_from_private_key_obj(key_api, private_key):
    signature = private_key.sign_msg_non_recoverable(MSG)

    assert key_api.ecdsa_verify(MSGHASH, signature, private_key.public_key)


def test_hash_signing_from_private_key_obj(key_api, private_key):
    signature = private_key.sign_msg_hash(MSGHASH)

    assert key_api.ecdsa_verify(MSGHASH, signature, private_key.public_key)


def test_hash_signing_non_recoverable_from_private_key_obj(key_api, private_key):
    signature = private_key.sign_msg_hash_non_recoverable(MSGHASH)

    assert key_api.ecdsa_verify(MSGHASH, signature, private_key.public_key)


def test_recover_from_public_key_class(key_api, private_key):
    signature = key_api.ecdsa_sign(MSGHASH, private_key)
    public_key = key_api.PublicKey.recover_from_msg_hash(MSGHASH, signature)

    assert public_key == key_api.PublicKey.recover_from_msg(MSG, signature)
    assert public_key == private_key.public_key


def test_verify_from_public_key_obj(key_api, private_key):
    non_recoverable_signature = key_api.ecdsa_sign_non_recoverable(MSGHASH, private_key)
    recoverable_signature = key_api.ecdsa_sign_non_recoverable(MSGHASH, private_key)

    public_key = private_key.public_key

    for signature in (recoverable_signature, non_recoverable_signature):
        assert public_key.verify_msg_hash(MSGHASH, signature)
        assert public_key.verify_msg(MSG, signature)


def test_from_private_for_public_key_class(key_api, private_key):
    public_key = key_api.PublicKey.from_private(private_key)

    assert public_key == private_key.public_key


def test_verify_from_signature_obj(key_api, private_key):
    signature = key_api.ecdsa_sign(MSGHASH, private_key)

    assert signature.verify_msg_hash(MSGHASH, private_key.public_key)
    assert signature.verify_msg(MSG, private_key.public_key)


def test_verify_from_non_recoverable_signature_obj(key_api, private_key):
    signature = key_api.ecdsa_sign(MSGHASH, private_key).to_non_recoverable_signature()

    assert signature.verify_msg_hash(MSGHASH, private_key.public_key)
    assert signature.verify_msg(MSG, private_key.public_key)


def test_recover_from_signature_obj(key_api, private_key):
    signature = key_api.ecdsa_sign(MSGHASH, private_key)
    public_key = signature.recover_public_key_from_msg_hash(MSGHASH)

    assert public_key == signature.recover_public_key_from_msg(MSG)
    assert public_key == private_key.public_key


def test_to_address_from_public_key(private_key):
    address = private_key.public_key.to_address()
    assert is_normalized_address(address)
    assert is_same_address(address, ADDRESS)


def test_to_checksum_address_from_public_key(private_key):
    address = private_key.public_key.to_checksum_address()
    assert is_checksum_address(address)
    assert is_same_address(address, ADDRESS)


def test_to_canonical_address_from_public_key(private_key):
    address = private_key.public_key.to_canonical_address()
    assert is_canonical_address(address)
    assert is_same_address(address, ADDRESS)


def test_hex_conversion(private_key):
    public_key = private_key.public_key
    signature = private_key.sign_msg(b"message")

    assert hex(public_key) == encode_hex(public_key.to_bytes())
    assert hex(private_key) == encode_hex(private_key.to_bytes())
    assert hex(signature) == encode_hex(signature.to_bytes())

    assert public_key.to_hex() == encode_hex(public_key.to_bytes())
    assert private_key.to_hex() == encode_hex(private_key.to_bytes())
    assert signature.to_hex() == encode_hex(signature.to_bytes())


def test_bytes_conversion(key_api, private_key):
    public_key = private_key.public_key
    signature = private_key.sign_msg(b"message")

    assert public_key.to_bytes() == public_key._raw_key
    assert private_key.to_bytes() == private_key._raw_key
    assert signature.to_bytes() == key_api.Signature(signature.to_bytes()).to_bytes()


def test_compressed_bytes_conversion(key_api, private_key):
    public_key = private_key.public_key
    compressed_bytes = public_key.to_compressed_bytes()
    assert len(compressed_bytes) == 33
    assert key_api.PublicKey.from_compressed_bytes(compressed_bytes) == public_key


@pytest.mark.parametrize(
    "validation_error", (ValidationError, EthKeysValidationErrorCopy)
)
def test_compressed_bytes_validation(key_api, private_key, validation_error):
    valid_key = private_key.public_key.to_compressed_bytes()

    with pytest.raises(validation_error):
        key_api.PublicKey.from_compressed_bytes(valid_key + b"\x00")
    with pytest.raises(validation_error):
        key_api.PublicKey.from_compressed_bytes(valid_key[:-1])
    with pytest.raises(validation_error):
        key_api.PublicKey.from_compressed_bytes(b"\x04" + valid_key[1:])


def test_validation_error_is_from_eth_utils():
    assert EthKeysValidationErrorCopy is ValidationError



from eth_keys.datatypes import (
    BaseSignature,
    NonRecoverableSignature,
    PrivateKey,
    PublicKey,
    Signature,
)


class BaseECCBackend:
    def ecdsa_sign(self, msg_hash: bytes, private_key: PrivateKey) -> Signature:
        raise NotImplementedError()

    def ecdsa_sign_non_recoverable(
        self, msg_hash: bytes, private_key: PrivateKey
    ) -> NonRecoverableSignature:
        raise NotImplementedError()

    def ecdsa_verify(
        self, msg_hash: bytes, signature: BaseSignature, public_key: PublicKey
    ) -> bool:
        raise NotImplementedError()

    def ecdsa_recover(self, msg_hash: bytes, signature: Signature) -> PublicKey:
        raise NotImplementedError()

    def private_key_to_public_key(self, private_key: PrivateKey) -> PublicKey:
        raise NotImplementedError()

    def decompress_public_key_bytes(self, compressed_public_key_bytes: bytes) -> bytes:
        raise NotImplementedError()

    def compress_public_key_bytes(self, uncompressed_public_key_bytes: bytes) -> bytes:
        raise NotImplementedError()



from typing import (
    Any,
)

from eth_utils import (
    ValidationError,
    encode_hex,
    is_bytes,
    is_integer,
)
from eth_utils.toolz import (
    curry,
)

from eth_keys.constants import (
    SECPK1_N,
)


def validate_integer(value: Any) -> None:
    if not is_integer(value) or isinstance(value, bool):
        raise ValidationError(f"Value must be a an integer.  Got: {type(value)}")


def validate_bytes(value: Any) -> None:
    if not is_bytes(value):
        raise ValidationError(f"Value must be a byte string.  Got: {type(value)}")


@curry
def validate_gte(value: Any, minimum: int) -> None:
    validate_integer(value)
    if value < minimum:
        raise ValidationError(
            f"Value {value} is not greater than or equal to {minimum}"
        )


@curry
def validate_lte(value: Any, maximum: int) -> None:
    validate_integer(value)
    if value > maximum:
        raise ValidationError(f"Value {value} is not less than or equal to {maximum}")


validate_lt_secpk1n = validate_lte(maximum=SECPK1_N - 1)


def validate_bytes_length(value: bytes, expected_length: int, name: str) -> None:
    actual_length = len(value)
    if actual_length != expected_length:
        raise ValidationError(
            f"Unexpected {name} length: Expected {expected_length}, but got "
            f"{actual_length} bytes"
        )


def validate_message_hash(value: Any) -> None:
    validate_bytes(value)
    validate_bytes_length(value, 32, "message hash")


def validate_uncompressed_public_key_bytes(value: Any) -> None:
    validate_bytes(value)
    validate_bytes_length(value, 64, "uncompressed public key")


def validate_compressed_public_key_bytes(value: Any) -> None:
    validate_bytes(value)
    validate_bytes_length(value, 33, "compressed public key")
    first_byte = value[0:1]
    if first_byte not in (b"\x02", b"\x03"):
        raise ValidationError(
            "Unexpected compressed public key format: Must start with 0x02 or 0x03, "
            f"but starts with {encode_hex(first_byte)}"
        )


def validate_private_key_bytes(value: Any) -> None:
    validate_bytes(value)
    validate_bytes_length(value, 32, "private key")


def validate_recoverable_signature_bytes(value: Any) -> None:
    validate_bytes(value)
    validate_bytes_length(value, 65, "recoverable signature")


def validate_non_recoverable_signature_bytes(value: Any) -> None:
    validate_bytes(value)
    validate_bytes_length(value, 64, "non recoverable signature")


def validate_signature_v(value: int) -> None:
    validate_integer(value)
    validate_gte(value, minimum=0)
    validate_lte(value, maximum=1)


def validate_signature_r_or_s(value: int) -> None:
    validate_integer(value)
    validate_gte(value, 0)
    validate_lt_secpk1n(value)



import os

from eth_utils import (
    keccak,
)
from hypothesis import (
    given,
)
import pytest
from strategies import (
    message_hash_st,
    private_key_st,
)

from eth_keys import (
    KeyAPI,
)
from eth_keys.backends import (
    CoinCurveECCBackend,
    NativeECCBackend,
)
from eth_keys.constants import (
    SECPK1_N,
)
from eth_keys.exceptions import (
    BadSignature,
)
from eth_keys.utils.numeric import (
    coerce_low_s,
)

MSG = b"message"
MSGHASH = keccak(MSG)


backends = [
    NativeECCBackend(),
]

try:
    pass

    backends.append(CoinCurveECCBackend())
except ImportError:
    if "REQUIRE_COINCURVE" in os.environ:
        raise


def backend_id_fn(backend):
    return type(backend).__name__


@pytest.fixture(params=backends, ids=backend_id_fn, scope="module")
def key_api(request):
    return KeyAPI(backend=request.param)


def test_ecdsa_sign(key_api, key_fixture):
    private_key = key_api.PrivateKey(key_fixture["privkey"])
    signature = key_api.ecdsa_sign(MSGHASH, private_key)

    assert key_api.ecdsa_verify(MSGHASH, signature, private_key.public_key)


def test_ecdsa_sign_non_recoverable(key_api, key_fixture):
    private_key = key_api.PrivateKey(key_fixture["privkey"])
    signature = key_api.ecdsa_sign_non_recoverable(MSGHASH, private_key)
    non_recoverable_signature = key_api.ecdsa_sign_non_recoverable(MSGHASH, private_key)
    assert non_recoverable_signature.r == signature.r
    assert non_recoverable_signature.s == signature.s

    assert key_api.ecdsa_verify(MSGHASH, signature, private_key.public_key)


def test_ecdsa_verify(key_api, key_fixture):
    signature = key_api.Signature(vrs=key_fixture["raw_sig"])
    public_key = key_api.PublicKey(key_fixture["pubkey"])

    assert key_api.ecdsa_verify(MSGHASH, signature, public_key)


def test_ecdsa_recover(key_api, key_fixture):
    signature = key_api.Signature(vrs=key_fixture["raw_sig"])
    public_key = key_api.PublicKey(key_fixture["pubkey"])

    assert key_api.ecdsa_recover(MSGHASH, signature) == public_key


@pytest.mark.parametrize(
    "v,r,s,msghash",
    (
        (
            # Test params from the ethereum/tests repo
            0,
            int(
                "0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798", 16
            ),
            int(
                "0x6b8d2c81b11b2d699528dde488dbdf2f94293d0d33c32e347f255fa4a6c1f0a9", 16
            ),
            bytes.fromhex(
                "6b8d2c81b11b2d699528dde488dbdf2f94293d0d33c32e347f255fa4a6c1f0a9"
            ),
        ),
        (
            # Test params from signing a message using the private key: '0x' + '00' * 32
            1,
            29836180350949573232951565573845061551093497675544587026406064656720118638890,  # noqa: E501
            36811185137926304485684021052401800813557229270017380070056792973957337676018,  # noqa: E501
            bytes.fromhex(
                "1476abb745d423bf09273f1afd887d951181d25adc66c4834a70491911b7f750"
            ),
        ),
    ),
)
def test_ecdsa_recover_identity_point(key_api, key_fixture, v, r, s, msghash):
    signature = key_api.Signature(vrs=(v, r, s))

    with pytest.raises(BadSignature):
        key_api.ecdsa_recover(msghash, signature)


def test_decompress_public_key_bytes(key_api, key_fixture):
    compressed = key_fixture["compressed_pubkey"]
    uncompressed = key_fixture["pubkey"]

    key_from_compressed = key_api.PublicKey.from_compressed_bytes(compressed)
    assert key_from_compressed.to_bytes() == uncompressed


def test_compress_public_key_bytes(key_api, key_fixture):
    uncompressed = key_fixture["pubkey"]
    compressed = key_fixture["compressed_pubkey"]

    key_from_uncompressed = key_api.PublicKey(uncompressed)
    assert key_from_uncompressed.to_compressed_bytes() == compressed


@given(
    private_key_bytes=private_key_st,
)
def test_compress_decompress_inversion(key_api, private_key_bytes):
    private_key = key_api.PrivateKey(private_key_bytes)

    original = private_key.public_key
    compressed_bytes = original.to_compressed_bytes()
    decompressed = key_api.PublicKey.from_compressed_bytes(compressed_bytes)
    assert decompressed == original


@given(
    private_key_bytes=private_key_st,
    message_hash=message_hash_st,
)
def test_signatures_with_high_s(key_api, private_key_bytes, message_hash):
    private_key = key_api.PrivateKey(private_key_bytes)
    low_s_signature = private_key.sign_msg_hash(message_hash)
    assert coerce_low_s(low_s_signature.s) == low_s_signature.s
    high_s = -low_s_signature.s % SECPK1_N
    assert coerce_low_s(high_s) == low_s_signature.s
    high_s_signature = key_api.Signature(
        vrs=(low_s_signature.v, low_s_signature.r, high_s)
    )
    assert key_api.ecdsa_verify(message_hash, high_s_signature, private_key.public_key)



from eth_keys import (
    keys,
)
from eth_keys.tools.factories import (
    PrivateKeyFactory,
    PublicKeyFactory,
)


def test_private_key_factory():
    actual = PrivateKeyFactory()
    assert actual == keys.PrivateKey(actual.to_bytes())


def test_public_key_factory():
    actual = PublicKeyFactory()
    assert actual == keys.PublicKey(actual.to_bytes())



import asn1tools
from hypothesis import (
    example,
    given,
    settings,
    strategies as st,
)
from pyasn1.codec.der import (
    decoder as pyasn1_decoder,
    encoder as pyasn1_encoder,
)
from pyasn1.type import (
    namedtype,
    univ,
)
import pytest

from eth_keys.utils.der import (
    two_int_sequence_decoder,
    two_int_sequence_encoder,
)

ASN1_ECDSA_SPEC_STRING = """\
ECDSASpec DEFINITIONS ::= BEGIN
      ECDSASignature ::= SEQUENCE {
         r   INTEGER,
         s   INTEGER
     }
END
"""
ASN1_SPEC = asn1tools.compile_string(ASN1_ECDSA_SPEC_STRING, "der")


def asn1tools_encode(r, s):
    return ASN1_SPEC.encode("ECDSASignature", {"r": r, "s": s})


def asn1tools_decode(encoded):
    decoded = ASN1_SPEC.decode("ECDSASignature", encoded)
    return decoded["r"], decoded["s"]


class TwoInts(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType("r", univ.Integer()),
        namedtype.NamedType("s", univ.Integer()),
    )


def pyasn1_encode(r, s):
    structured = TwoInts()
    structured["r"] = r
    structured["s"] = s
    return pyasn1_encoder.encode(structured)


def pyasn1_decode(encoded):
    decoded = pyasn1_decoder.decode(encoded, asn1Spec=TwoInts())
    return decoded[0]["r"], decoded[0]["s"]


MAX_32_BYTE_INT = 256**32 - 1
uint32strategy = st.integers(min_value=0, max_value=MAX_32_BYTE_INT)


@pytest.mark.parametrize(
    "encoder, decoder",
    (
        (two_int_sequence_encoder, asn1tools_decode),
        (two_int_sequence_encoder, pyasn1_decode),
        (two_int_sequence_encoder, two_int_sequence_decoder),
        (asn1tools_encode, two_int_sequence_decoder),
        (pyasn1_encode, two_int_sequence_decoder),
    ),
    ids=(
        "local_encode=>asn1tools_decode",
        "local_encode=>pyasn1_decode",
        "local_encode=>local_decode",
        "asn1tools_encode=>local_decode",
        "pyasn1_encode=>local_decode",
    ),
)
@given(
    uint32strategy,
    uint32strategy,
)
@example(0, 0)
@example(MAX_32_BYTE_INT, MAX_32_BYTE_INT)
@example(MAX_32_BYTE_INT // 2, MAX_32_BYTE_INT // 2)
@example(MAX_32_BYTE_INT // 2 + 1, MAX_32_BYTE_INT // 2 + 1)
@settings(max_examples=500)
def test_encode_decode_pairings(encoder, decoder, r, s):
    encoded = encoder(r, s)
    end_r, end_s = decoder(encoded)
    assert (end_r, end_s) == (r, s)



from eth_utils import (
    ValidationError,
)

from eth_keys.datatypes import (
    BaseSignature,
    LazyBackend,
    NonRecoverableSignature,
    PrivateKey,
    PublicKey,
    Signature,
)
from eth_keys.validation import (
    validate_message_hash,
)

# These must be aliased due to a scoping issue in mypy
# https://github.com/python/mypy/issues/1775
_PublicKey = PublicKey
_PrivateKey = PrivateKey
_Signature = Signature
_NonRecoverableSignature = NonRecoverableSignature


class KeyAPI(LazyBackend):
    PublicKey = PublicKey
    PrivateKey = PrivateKey
    Signature = Signature
    NonRecoverableSignature = NonRecoverableSignature

    #
    # Proxy method calls to the backends
    #
    def ecdsa_sign(self, message_hash: bytes, private_key: _PrivateKey) -> _Signature:
        validate_message_hash(message_hash)
        if not isinstance(private_key, PrivateKey):
            raise ValidationError(
                "The `private_key` must be an instance of "
                "`eth_keys.datatypes.PrivateKey`"
            )
        signature = self.backend.ecdsa_sign(message_hash, private_key)
        if not isinstance(signature, Signature):
            raise ValidationError(
                "Backend returned an invalid signature.  Return value must be "
                "an instance of `eth_keys.datatypes.Signature`"
            )
        return signature

    def ecdsa_sign_non_recoverable(
        self, message_hash: bytes, private_key: _PrivateKey
    ) -> _NonRecoverableSignature:
        validate_message_hash(message_hash)
        if not isinstance(private_key, PrivateKey):
            raise ValidationError(
                "The `private_key` must be an instance of "
                "`eth_keys.datatypes.PrivateKey`"
            )
        signature = self.backend.ecdsa_sign_non_recoverable(message_hash, private_key)
        if not isinstance(signature, NonRecoverableSignature):
            raise ValidationError(
                "Backend returned an invalid signature.  Return value must be "
                "an instance of `eth_keys.datatypes.Signature`"
            )
        return signature

    def ecdsa_verify(
        self, message_hash: bytes, signature: BaseSignature, public_key: _PublicKey
    ) -> bool:
        validate_message_hash(message_hash)
        if not isinstance(public_key, PublicKey):
            raise ValidationError(
                "The `public_key` must be an instance of `eth_keys.datatypes.PublicKey`"
            )
        if not isinstance(signature, BaseSignature):
            raise ValidationError(
                "The `signature` must be an instance of "
                "`eth_keys.datatypes.BaseSignature`"
            )
        return self.backend.ecdsa_verify(message_hash, signature, public_key)

    def ecdsa_recover(self, message_hash: bytes, signature: _Signature) -> _PublicKey:
        validate_message_hash(message_hash)
        if not isinstance(signature, Signature):
            raise ValidationError(
                "The `signature` must be an instance of `eth_keys.datatypes.Signature`"
            )
        public_key = self.backend.ecdsa_recover(message_hash, signature)
        if not isinstance(public_key, _PublicKey):
            raise ValidationError(
                "Backend returned an invalid public_key.  Return value must be "
                "an instance of `eth_keys.datatypes.PublicKey`"
            )
        return public_key

    def private_key_to_public_key(self, private_key: _PrivateKey) -> _PublicKey:
        if not isinstance(private_key, PrivateKey):
            raise ValidationError(
                "The `private_key` must be an instance of "
                "`eth_keys.datatypes.PrivateKey`"
            )
        public_key = self.backend.private_key_to_public_key(private_key)
        if not isinstance(public_key, PublicKey):
            raise ValidationError(
                "Backend returned an invalid public_key.  Return value must be "
                "an instance of `eth_keys.datatypes.PublicKey`"
            )
        return public_key


# This creates an easy to import backend which will lazily fetch whatever
# backend has been configured at runtime (as opposed to import or instantiation time).
lazy_key_api = KeyAPI(backend=None)



from eth_utils import (
    ValidationError,
    keccak,
)
import pytest

from eth_keys import (
    KeyAPI,
)
from eth_keys.backends import (
    NativeECCBackend,
)

MSG = b"message"
MSGHASH = keccak(MSG)
PK_BYTES = b"\x01" * 32


@pytest.fixture
def ecc_backend():
    return NativeECCBackend()


@pytest.fixture
def key_api(ecc_backend):
    return KeyAPI(ecc_backend)


@pytest.fixture
def private_key(key_api):
    return key_api.PrivateKey(PK_BYTES)


@pytest.fixture
def public_key(private_key):
    return private_key.public_key


@pytest.fixture
def signature(private_key):
    return private_key.sign_msg_hash(MSGHASH)


def test_key_api_ecdsa_sign_validation(key_api, private_key):
    with pytest.raises(ValidationError):
        key_api.ecdsa_sign(MSGHASH, private_key.to_bytes())
    with pytest.raises(ValidationError):
        key_api.ecdsa_sign(MSG, private_key)

    signature = key_api.ecdsa_sign(MSGHASH, private_key)
    assert signature.verify_msg_hash(MSGHASH, private_key.public_key)


def test_key_api_ecdsa_sign_non_recoverable_validation(key_api, private_key):
    with pytest.raises(ValidationError):
        key_api.ecdsa_sign_non_recoverable(MSGHASH, private_key.to_bytes())
    with pytest.raises(ValidationError):
        key_api.ecdsa_sign_non_recoverable(MSG, private_key)

    signature = key_api.ecdsa_sign_non_recoverable(MSGHASH, private_key)
    assert signature.verify_msg_hash(MSGHASH, private_key.public_key)


def test_key_api_ecdsa_verify_validation(key_api, signature, public_key):
    with pytest.raises(ValidationError):
        key_api.ecdsa_verify(MSGHASH, signature.to_bytes(), public_key)
    with pytest.raises(ValidationError):
        key_api.ecdsa_verify(MSGHASH, signature, public_key.to_bytes())
    with pytest.raises(ValidationError):
        key_api.ecdsa_verify(MSG, signature, public_key)

    assert key_api.ecdsa_verify(MSGHASH, signature, public_key)


def test_key_api_ecdsa_recover_validation(key_api, signature, public_key):
    with pytest.raises(ValidationError):
        key_api.ecdsa_recover(MSGHASH, signature.to_bytes())
    with pytest.raises(ValidationError):
        key_api.ecdsa_recover(MSG, signature)

    assert key_api.ecdsa_recover(MSGHASH, signature) == public_key



from eth_utils import (
    keccak,
)
from hypothesis import (
    given,
    settings,
    strategies as st,
)
import pytest
from strategies import (
    message_hash_st,
    private_key_st,
    signature_st,
)

from eth_keys import (
    KeyAPI,
)
from eth_keys.backends import (
    CoinCurveECCBackend,
    NativeECCBackend,
)
from eth_keys.exceptions import (
    BadSignature,
)

MSG = b"message"
MSGHASH = keccak(MSG)

MAX_EXAMPLES = 200


@pytest.fixture(scope="module")
def native_key_api():
    return KeyAPI(backend=NativeECCBackend())


@pytest.fixture(scope="module")
def coincurve_key_api():
    return KeyAPI(backend=CoinCurveECCBackend())


@given(
    private_key_bytes=private_key_st,
)
@settings(max_examples=MAX_EXAMPLES)
def test_public_key_generation_is_equal(
    private_key_bytes, native_key_api, coincurve_key_api
):
    native_public_key = native_key_api.PrivateKey(private_key_bytes).public_key
    coincurve_public_key = coincurve_key_api.PrivateKey(private_key_bytes).public_key

    assert native_public_key == coincurve_public_key


@given(
    private_key_bytes=private_key_st,
    message_hash=message_hash_st,
)
@settings(max_examples=MAX_EXAMPLES)
def test_signing_is_equal(
    private_key_bytes, message_hash, native_key_api, coincurve_key_api
):
    native_private_key = native_key_api.PrivateKey(private_key_bytes)
    native_signature = native_key_api.ecdsa_sign(message_hash, native_private_key)
    native_non_recoverable_signature = native_key_api.ecdsa_sign_non_recoverable(
        message_hash,
        native_private_key,
    )

    coincurve_private_key = coincurve_key_api.PrivateKey(private_key_bytes)
    coincurve_signature = coincurve_key_api.ecdsa_sign(
        message_hash, coincurve_private_key
    )
    coincurve_non_recoverable_signature = coincurve_key_api.ecdsa_sign_non_recoverable(
        message_hash,
        coincurve_private_key,
    )

    assert native_signature == coincurve_signature
    assert native_non_recoverable_signature == coincurve_non_recoverable_signature


@given(
    private_key_bytes=private_key_st,
    message_hash=message_hash_st,
    direction=st.one_of(
        st.just("coincurve-to-native"),
        st.just("native-to-coincurve"),
    ),
)
@settings(max_examples=MAX_EXAMPLES)
def test_native_to_coincurve_recover(
    private_key_bytes, message_hash, direction, native_key_api, coincurve_key_api
):
    if direction == "coincurve-to-native":
        backend_a = coincurve_key_api
        backend_b = native_key_api
    elif direction == "native-to-coincurve":
        backend_b = coincurve_key_api
        backend_a = native_key_api
    else:
        raise AssertionError("invariant")

    private_key_a = backend_a.PrivateKey(private_key_bytes)
    public_key_a = private_key_a.public_key
    signature_a = backend_a.ecdsa_sign(message_hash, private_key_a)

    public_key_b = backend_b.ecdsa_recover(message_hash, signature_a)
    assert public_key_b == public_key_a
    assert backend_b.ecdsa_verify(message_hash, signature_a, public_key_b)


@given(
    message_hash=message_hash_st,
    signature_bytes=signature_st,
    direction=st.one_of(
        st.just("coincurve-to-native"),
        st.just("native-to-coincurve"),
    ),
)
@settings(max_examples=MAX_EXAMPLES)
def test_coincurve_to_native_invalid_signatures(
    message_hash, signature_bytes, direction, native_key_api, coincurve_key_api
):
    if direction == "coincurve-to-native":
        backend_a = coincurve_key_api
        backend_b = native_key_api
    elif direction == "native-to-coincurve":
        backend_b = coincurve_key_api
        backend_a = native_key_api
    else:
        raise AssertionError("invariant")

    try:
        signature_a = backend_a.Signature(signature_bytes)
    except BadSignature:
        is_bad_signature = True
    else:
        is_bad_signature = False

    if is_bad_signature:
        with pytest.raises(BadSignature):
            backend_b.Signature(signature_bytes)
        return
    try:
        public_key_a = backend_a.ecdsa_recover(message_hash, signature_a)
    except BadSignature:
        is_bad_recovery = True
    else:
        is_bad_recovery = False

    if is_bad_recovery:
        with pytest.raises(BadSignature):
            backend_b.ecdsa_recover(message_hash, signature_a)
        return

    public_key_b = backend_b.ecdsa_recover(message_hash, signature_a)

    assert public_key_b == public_key_a


@given(
    private_key_bytes=private_key_st,
)
def test_public_key_compression_is_equal(
    private_key_bytes, native_key_api, coincurve_key_api
):
    native_public_key = native_key_api.PrivateKey(private_key_bytes).public_key
    coincurve_public_key = coincurve_key_api.PrivateKey(private_key_bytes).public_key

    native_compressed_public_key = native_public_key.to_compressed_bytes()
    coincurve_compressed_public_key = coincurve_public_key.to_compressed_bytes()

    assert native_compressed_public_key == coincurve_compressed_public_key


@given(
    private_key_bytes=private_key_st,
)
def test_public_key_decompression_is_equal(
    private_key_bytes, native_key_api, coincurve_key_api
):
    public_key_template = coincurve_key_api.PrivateKey(private_key_bytes).public_key
    compressed_public_key = public_key_template.to_compressed_bytes()

    native_public_key = native_key_api.PublicKey.from_compressed_bytes(
        compressed_public_key
    )
    coincurve_public_key = coincurve_key_api.PublicKey.from_compressed_bytes(
        compressed_public_key
    )

    assert native_public_key == coincurve_public_key



#!/usr/bin/env python
from setuptools import (
    find_packages,
    setup,
)

extras_require = {
    "coincurve": [
        "coincurve>=12.0.0",
    ],
    "dev": [
        "build>=0.9.0",
        "bumpversion>=0.5.3",
        "ipython",
        "pre-commit>=3.4.0",
        "tox>=4.0.0",
        "twine",
        "wheel",
    ],
    "docs": [
        "towncrier>=21,<22",
    ],
    "test": [
        "pytest>=7.0.0",
        "asn1tools>=0.146.2",
        "factory-boy>=3.0.1",
        "pyasn1>=0.4.5",
        "hypothesis>=5.10.3,<6",
        "eth-hash[pysha3]",
    ],
}

extras_require["dev"] = (
    extras_require["coincurve"]
    + extras_require["dev"]
    + extras_require["docs"]
    + extras_require["test"]
)


with open("./README.md") as readme:
    long_description = readme.read()


setup(
    name="eth-keys",
    # *IMPORTANT*: Don't manually change the version here. Use `make bump`, as described in readme
    version="0.5.0",
    description="""eth-keys: Common API for Ethereum key operations""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="The Ethereum Foundation",
    author_email="snakecharmers@ethereum.org",
    url="https://github.com/ethereum/eth-keys",
    include_package_data=True,
    install_requires=[
        "eth-utils>=2",
        "eth-typing>=3",
    ],
    python_requires=">=3.8, <4",
    extras_require=extras_require,
    py_modules=["eth_keys"],
    license="MIT",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"eth_keys": ["py.typed"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



from eth_utils import (
    decode_hex,
    keccak,
)
import pytest

MSG = b"message"
MSGHASH = keccak(MSG)


# This is a sample of signatures generated with a known-good implementation of the ECDSA
# algorithm, which we use to test our ECC backends. If necessary, it can be generated
# from scratch with the following code:
"""
from devp2p import crypto
from eth_utils import encode_hex
msg = b'message'
msghash = crypto.sha3(b'message')
for secret in ['alice', 'bob', 'eve']:
    print("'{}': dict(".format(secret))
    privkey = crypto.mk_privkey(secret)
    pubkey = crypto.privtopub(privkey)
    print("    privkey='{}',".format(encode_hex(privkey)))
    print("    pubkey='{}',".format(encode_hex(crypto.privtopub(privkey))))
    ecc = crypto.ECCx(raw_privkey=privkey)
    sig = ecc.sign(msghash)
    print("    sig='{}',".format(encode_hex(sig)))
    print("    raw_sig='{}')".format(crypto._decode_sig(sig)))
    assert crypto.ecdsa_recover(msghash, sig) == pubkey
"""
# Compressed public keys have been calculated with coincurve


SECRETS = {
    "alice": dict(
        privkey=decode_hex(
            "0x9c0257114eb9399a2985f8e75dad7600c5d89fe3824ffa99ec1c3eb8bf3b0501"
        ),
        pubkey=decode_hex(
            "0x5eed5fa3a67696c334762bb4823e585e2ee579aba3558d9955296d6c04541b426078dbd48d74af1fd0c72aa1a05147cf17be6b60bdbed6ba19b08ec28445b0ca"  # noqa: E501
        ),
        compressed_pubkey=decode_hex(
            "0x025eed5fa3a67696c334762bb4823e585e2ee579aba3558d9955296d6c04541b42"
        ),
        sig=decode_hex(
            "0xb20e2ea5d3cbaa83c1e0372f110cf12535648613b479b64c1a8c1a20c5021f380434d07ec5795e3f789794351658e80b7faf47a46328f41e019d7b853745cdfd01"  # noqa: E501
        ),
        raw_sig=(
            1,
            80536744857756143861726945576089915884233437828013729338039544043241440681784,  # noqa: E501
            1902566422691403459035240420865094128779958320521066670269403689808757640701,  # noqa: E501
        ),
    ),
    "bob": dict(
        privkey=decode_hex(
            "0x38e47a7b719dce63662aeaf43440326f551b8a7ee198cee35cb5d517f2d296a2"
        ),
        pubkey=decode_hex(
            "0x347746ccb908e583927285fa4bd202f08e2f82f09c920233d89c47c79e48f937d049130e3d1c14cf7b21afefc057f71da73dec8e8ff74ff47dc6a574ccd5d570"  # noqa: E501
        ),
        compressed_pubkey=decode_hex(
            "0x02347746ccb908e583927285fa4bd202f08e2f82f09c920233d89c47c79e48f937"
        ),
        sig=decode_hex(
            "0x5c48ea4f0f2257fa23bd25e6fcb0b75bbe2ff9bbda0167118dab2bb6e31ba76e691dbdaf2a231fc9958cd8edd99507121f8184042e075cf10f98ba88abff1f3601"  # noqa: E501
        ),
        raw_sig=(
            1,
            41741612198399299636429810387160790514780876799439767175315078161978521003886,  # noqa: E501
            47545396818609319588074484786899049290652725314938191835667190243225814114102,  # noqa: E501
        ),
    ),
    "eve": dict(
        privkey=decode_hex(
            "0x876be0999ed9b7fc26f1b270903ef7b0c35291f89407903270fea611c85f515c"
        ),
        pubkey=decode_hex(
            "0xc06641f0d04f64dba13eac9e52999f2d10a1ff0ca68975716b6583dee0318d91e7c2aed363ed22edeba2215b03f6237184833fd7d4ad65f75c2c1d5ea0abecc0"  # noqa: E501
        ),
        compressed_pubkey=decode_hex(
            "0x02c06641f0d04f64dba13eac9e52999f2d10a1ff0ca68975716b6583dee0318d91"
        ),
        sig=decode_hex(
            "0xbabeefc5082d3ca2e0bc80532ab38f9cfb196fb9977401b2f6a98061f15ed603603d0af084bf906b2cdf6cdde8b2e1c3e51a41af5e9adec7f3643b3f1aa2aadf00"  # noqa: E501
        ),
        raw_sig=(
            0,
            84467545608142925331782333363288012579669270632210954476013542647119929595395,  # noqa: E501
            43529886636775750164425297556346136250671451061152161143648812009114516499167,  # noqa: E501
        ),
    ),
}


@pytest.fixture(params=["alice", "bob", "eve"])
def key_fixture(request):
    if request.param == "alice":
        return SECRETS["alice"]
    elif request.param == "bob":
        return SECRETS["bob"]
    elif request.param == "eve":
        return SECRETS["eve"]
    else:
        raise AssertionError("Should be unreachable")
